@@ -3,23 +3,23 @@
 # helper function for adjusting matrix dimensions and creating an offset matrix
 tergmprepare <- function(formula, offset = TRUE, blockdiag = FALSE, 
     verbose = TRUE) {
-  
   # extract response networks and make sure they are saved in a list
-  env <- new.env()
-  env$lhs.original <- deparse(formula[[2]])  # for reporting purposes later on
-  env$networks <- eval(parse(text = deparse(formula[[2]])))
-  if (class(env$networks) == "list" || class(env$networks) == "network.list") {
+  l <- list()
+  l$lhs.original <- deparse(formula[[2]])  # for reporting purposes later on
+  l$networks <- eval(parse(text = deparse(formula[[2]])), 
+      envir = environment(formula))
+  if (class(l$networks) == "list" || class(l$networks) == "network.list") {
     # do nothing
   } else {
-    env$networks <- list(env$networks)
+    l$networks <- list(l$networks)
   }
   
   # convert list elements to matrices if unknown data type
-  for (i in 1:length(env$networks)) {
-    if (!class(env$networks[[i]]) %in% c("network", "matrix", "list")) {
+  for (i in 1:length(l$networks)) {
+    if (!class(l$networks[[i]]) %in% c("network", "matrix", "list")) {
       tryCatch(
         {
-          env$networks[[i]] <- as.matrix(env$networks[[i]])
+          l$networks[[i]] <- as.matrix(l$networks[[i]])
         }, 
         error = function(cond) {
           stop(paste("Object", i, "could not be converted to a matrix."))
@@ -29,34 +29,34 @@ tergmprepare <- function(formula, offset = TRUE, blockdiag = FALSE,
   }
   
   # extract additional information
-  env$num.vertices <- max(sapply(env$networks, function(x) 
+  l$num.vertices <- max(sapply(l$networks, function(x) 
       network::get.network.attribute(network::network(x), "n")))  # number nodes
-  if (is.network(env$networks[[1]])) {
-    env$directed <- network::is.directed(env$networks[[1]])  # directed?
-    env$bipartite <- network::is.bipartite(env$networks[[1]])  # bipartite?
+  if (is.network(l$networks[[1]])) {
+    l$directed <- network::is.directed(l$networks[[1]])  # directed?
+    l$bipartite <- network::is.bipartite(l$networks[[1]])  # bipartite?
   } else {
-    if (xergm.common::is.mat.directed(as.matrix(env$networks[[1]]))) {
-      env$directed <- TRUE
+    if (xergm.common::is.mat.directed(as.matrix(l$networks[[1]]))) {
+      l$directed <- TRUE
     } else {
-      env$directed <- FALSE
+      l$directed <- FALSE
     }
-    if (xergm.common::is.mat.onemode(as.matrix(env$networks[[1]]))) {
-      env$bipartite <- FALSE
+    if (xergm.common::is.mat.onemode(as.matrix(l$networks[[1]]))) {
+      l$bipartite <- FALSE
     } else {
-      env$bipartite <- TRUE
+      l$bipartite <- TRUE
     }
   }
   
   # adjust and disassemble formula
-  env$form <- update.formula(formula, networks[[i]] ~ .)
-  env$time.steps <- length(env$networks)
-  tilde <- deparse(env$form[[1]])
-  lhs <- deparse(env$form[[2]])
-  rhs <- paste(deparse(env$form[[3]]), collapse = "")  # for long formulae
+  l$form <- update.formula(formula, networks[[i]] ~ .)
+  l$time.steps <- length(l$networks)
+  tilde <- deparse(l$form[[1]])
+  lhs <- deparse(l$form[[2]])
+  rhs <- paste(deparse(l$form[[3]]), collapse = "")  # for long formulae
   rhs <- gsub("\\s+", " ", rhs)
   
   # parse rhs of formula and add indices to edgecov and dyadcov terms
-  env$rhs.terms <- strsplit(rhs, "\\s*(\\+|\\*)\\s*")[[1]]
+  l$rhs.terms <- strsplit(rhs, "\\s*(\\+|\\*)\\s*")[[1]]
   rhs.indices <- gregexpr("\\+|\\*", rhs)[[1]]
   if (length(rhs.indices) == 1 && rhs.indices < 0) {
     rhs.operators <- character()
@@ -66,29 +66,29 @@ tergmprepare <- function(formula, offset = TRUE, blockdiag = FALSE,
   
   # preprocess dyadcov and edgecov terms, memory terms, and timecov terms
   covnames <- character()
-  for (k in 1:length(env$rhs.terms)) {
-    if (grepl("((edge)|(dyad))cov", env$rhs.terms[k])) {  # edgecov or dyadcov
+  for (k in 1:length(l$rhs.terms)) {
+    if (grepl("((edge)|(dyad))cov", l$rhs.terms[k])) {  # edgecov or dyadcov
       
       # split up into components
-      if (grepl(",\\s*?((attr)|\\\")", env$rhs.terms[k])) { # with attrib arg.
+      if (grepl(",\\s*?((attr)|\\\")", l$rhs.terms[k])) { # with attrib arg.
         s <- "((?:offset\\()?((edge)|(dyad))cov\\()([^\\)]+)((,\\s*a*.*?)\\)(?:\\))?)"
       } else { # without attribute argument
         s <- "((?:offset\\()?((edge)|(dyad))cov\\()([^\\)]+)((,*\\s*a*.*?)\\)(?:\\))?)"
       }
-      x1 <- sub(s, "\\1", env$rhs.terms[k], perl = TRUE)  # before the covariate
-      x2 <- sub(s, "\\5", env$rhs.terms[k], perl = TRUE)  # name of the cov.
+      x1 <- sub(s, "\\1", l$rhs.terms[k], perl = TRUE)  # before the covariate
+      x2 <- sub(s, "\\5", l$rhs.terms[k], perl = TRUE)  # name of the cov.
       if (grepl("\\[.*\\]", x2)) {
        stop(paste0("Covariate names are not allowed to have indices: ", x2, 
            ". Please prepare a list object before estimation."))
       }
       if (grepl("^\"", x2)) next  # ignore built-in matrices b/c conformable
-      x3 <- sub(s, "\\6", env$rhs.terms[k], perl = TRUE)  # after the covariate
-      x.current <- eval(parse(text = x2))
+      x3 <- sub(s, "\\6", l$rhs.terms[k], perl = TRUE)  # after the covariate
+      x.current <- eval(parse(text = x2), envir = environment(formula))
       type <- class(x.current)
-      env$covnames <- c(env$covnames, x2)
-      env[[x2]] <- x.current
+      l$covnames <- c(l$covnames, x2)
+      l[[x2]] <- x.current
       if (grepl("\\[i\\]+$", x2)) {
-        stop(paste0("Error in the following model term: ", env$rhs.terms[k], 
+        stop(paste0("Error in the following model term: ", l$rhs.terms[k], 
             ". The index 'i' is used internally by btergm. Please use a ", 
             "different index, for example 'j'."))
       }
@@ -96,15 +96,15 @@ tergmprepare <- function(formula, offset = TRUE, blockdiag = FALSE,
       # add brackets if necessary, convert to list, and reassemble rhs term
       if (grepl("[^\\]]\\]$", x2)) {
         # time-varying covariate with given indices (e.g., formula[1:5])
-        env$rhs.terms[k] <- paste0(x1, x2, x3)
+        l$rhs.terms[k] <- paste0(x1, x2, x3)
         if (type %in% c("matrix", "network", "dgCMatrix", "dgTMatrix", 
             "dsCMatrix", "dsTMatrix", "dgeMatrix")) {
           x.current <-list(x.current)
-          env[[x2]] <- x.current
+          l[[x2]] <- x.current
         }
-        if (length(x.current) != env$time.steps) {
+        if (length(x.current) != l$time.steps) {
           stop(paste(x2, "has", length(x.current), "elements, but there are", 
-              env$time.steps, "networks to be modeled."))
+              l$time.steps, "networks to be modeled."))
         }
         if (blockdiag == TRUE) {
           # do not add brackets
@@ -117,32 +117,32 @@ tergmprepare <- function(formula, offset = TRUE, blockdiag = FALSE,
         if (!type %in% c("matrix", "network")) {
           x.current <- as.matrix(x.current)
         }
-        env[[x2]] <- list()
-        for (i in 1:env$time.steps) {
-          env[[x2]][[i]] <- x.current
+        l[[x2]] <- list()
+        for (i in 1:l$time.steps) {
+          l[[x2]][[i]] <- x.current
         }
         if (blockdiag == TRUE) {
           # do not add brackets
         } else {
           x2 <- paste0(x2, "[[i]]")
         }
-        env$rhs.terms[k] <- paste(x1, x2, x3, sep = "")
+        l$rhs.terms[k] <- paste(x1, x2, x3, sep = "")
       } else if (type == "list" || type == "network.list") {
         # time-varying covariate
-        if (length(x.current) != env$time.steps) {
+        if (length(x.current) != l$time.steps) {
           stop(paste(x2, "has", length(get(x2)), "elements, but there are", 
-              env$time.steps, "networks to be modeled."))
+              l$time.steps, "networks to be modeled."))
         }
         if (blockdiag == TRUE) {
           # do not add brackets
         } else {
           x2 <- paste0(x2, "[[i]]")
         }
-        env$rhs.terms[k] <- paste0(x1, x2, x3)
+        l$rhs.terms[k] <- paste0(x1, x2, x3)
       } else {  # something else --> try to convert to matrix list
         tryCatch(
           {
-            env[[x2]] <- list(rep(as.matrix(x.current)), env$time.steps)
+            l[[x2]] <- list(rep(as.matrix(x.current)), l$time.steps)
           }, 
           error = function(cond) {
             stop(paste0("Object '", x2, 
@@ -150,30 +150,29 @@ tergmprepare <- function(formula, offset = TRUE, blockdiag = FALSE,
           }
         )
       }
-    } else if (grepl("memory", env$rhs.terms[k])) {  # memory terms
+    } else if (grepl("memory", l$rhs.terms[k])) {  # memory terms
       
       # extract type argument
       s <- "(?:memory\\((?:.*type\\s*=\\s*)?(?:\"|'))(\\w+)(?:(\"|').*\\))"
-      if (grepl(s, env$rhs.terms[k]) == FALSE) {
+      if (grepl(s, l$rhs.terms[k]) == FALSE) {
         type <- "stability"
       } else {
-        type <- sub(s, "\\1", env$rhs.terms[k], perl = TRUE)
+        type <- sub(s, "\\1", l$rhs.terms[k], perl = TRUE)
       }
       
       # extract lag argument
       s <- "(?:memory\\(.*lag\\s*=\\s*)(\\d+)(?:.*\\))"
-      if (grepl(s, env$rhs.terms[k]) == FALSE) {
+      if (grepl(s, l$rhs.terms[k]) == FALSE) {
         lag <- 1
       } else {
-        lag <- as.integer(sub(s, "\\1", env$rhs.terms[k], perl = TRUE))
+        lag <- as.integer(sub(s, "\\1", l$rhs.terms[k], perl = TRUE))
       }
-      if (lag > length(env$networks) - 1) {
+      if (lag > length(l$networks) - 1) {
         stop("The 'lag' argument in the 'memory' term is too large.")
       }
       
       # process dependent list of networks
-      mem <- env$networks[-(length(env$networks):(length(env$networks) - lag + 
-          1))]
+      mem <- l$networks[-(length(l$networks):(length(l$networks) - lag + 1))]
       mem <- lapply(mem, as.matrix)
       memory <- list()
       for (i in 1:length(mem)) {
@@ -197,37 +196,36 @@ tergmprepare <- function(formula, offset = TRUE, blockdiag = FALSE,
       rm(mem)
       
       # re-introduce as edgecov and name of model term including brackets
-      env[["memory"]] <- memory
+      l[["memory"]] <- memory
       if (blockdiag == TRUE) {
-        env$rhs.terms[k] <- "edgecov(memory)"
+        l$rhs.terms[k] <- "edgecov(memory)"
       } else {
-        env$rhs.terms[k] <- "edgecov(memory[[i]])"
+        l$rhs.terms[k] <- "edgecov(memory[[i]])"
       }
-      env$covnames <- c(env$covnames, "memory")
-    } else if (grepl("delrecip", env$rhs.terms[k])) {  # delayed reciprocity
+      l$covnames <- c(l$covnames, "memory")
+    } else if (grepl("delrecip", l$rhs.terms[k])) {  # delayed reciprocity
       
       # extract mutuality argument
       s <- "(?:delrecip\\((?:.*mutuality\\s*=\\s*)?)((TRUE)|(FALSE)|T|F)(?:.*\\))"
-      if (grepl(s, env$rhs.terms[k]) == FALSE) {
+      if (grepl(s, l$rhs.terms[k]) == FALSE) {
         mutuality <- FALSE
       } else {
-        mutuality <- as.logical(sub(s, "\\1", env$rhs.terms[k], perl = TRUE))
+        mutuality <- as.logical(sub(s, "\\1", l$rhs.terms[k], perl = TRUE))
       }
       
       # extract lag argument
       s <- "(?:delrecip\\(.*lag\\s*=\\s*)(\\d+)(?:.*\\))"  # get lag
-      if (grepl(s, env$rhs.terms[k]) == FALSE) {
+      if (grepl(s, l$rhs.terms[k]) == FALSE) {
         lag <- 1
       } else {
-        lag <- as.integer(sub(s, "\\1", env$rhs.terms[k], perl = TRUE))
+        lag <- as.integer(sub(s, "\\1", l$rhs.terms[k], perl = TRUE))
       }
-      if (lag > length(env$networks) - 1) {
+      if (lag > length(l$networks) - 1) {
         stop("The 'lag' argument in the 'delrecip' term is too large.")
       }
       
       # process dependent list of networks
-      dlr <- env$networks[-(length(env$networks):(length(env$networks) - lag + 
-          1))]
+      dlr <- l$networks[-(length(l$networks):(length(l$networks) - lag + 1))]
       dlr <- lapply(dlr, function(x) t(as.matrix(x)))
       delrecip <- list()
       for (i in 1:length(dlr)) {
@@ -239,58 +237,58 @@ tergmprepare <- function(formula, offset = TRUE, blockdiag = FALSE,
       rm(dlr)
       
       # re-introduce as edgecov and name of model term including brackets
-      env[["delrecip"]] <- delrecip
+      l[["delrecip"]] <- delrecip
       if (blockdiag == TRUE) {
-        env$rhs.terms[k] <- "edgecov(delrecip)"
+        l$rhs.terms[k] <- "edgecov(delrecip)"
       } else {
-        env$rhs.terms[k] <- "edgecov(delrecip[[i]])"
+        l$rhs.terms[k] <- "edgecov(delrecip[[i]])"
       }
-      env$covnames <- c(env$covnames, "delrecip")
-    } else if (grepl("timecov", env$rhs.terms[k])) {  # time covariate
+      l$covnames <- c(l$covnames, "delrecip")
+    } else if (grepl("timecov", l$rhs.terms[k])) {  # time covariate
       
       # extract x argument
       s <- "(?:timecov\\((?:.*x\\s*=\\s*)?)(\\w+)(?:.*\\))"
-      if (sub(s, "\\1", env$rhs.terms[k], perl = TRUE) %in% c("minimum", 
+      if (sub(s, "\\1", l$rhs.terms[k], perl = TRUE) %in% c("minimum", 
           "maximum", "transform", "min", "max", "trans")) {
         s <- "(?:timecov\\(?:.*x\\s*=\\s*)(\\w+)(?:.*\\))"
       }
-      if (grepl(s, env$rhs.terms[k]) == FALSE) {
+      if (grepl(s, l$rhs.terms[k]) == FALSE) {
         x <- NULL
         suffix <- ""
         label <- "timecov"
       } else {
-        x <- sub(s, "\\1", env$rhs.terms[k], perl = TRUE)
+        x <- sub(s, "\\1", l$rhs.terms[k], perl = TRUE)
         label <- paste0("timecov.", x)
       }
       
       # extract minimum argument
       s <- "(?:timecov\\(.*minimum\\s*=\\s*)(\\d+)(?:.*\\))"
-      if (grepl(s, env$rhs.terms[k]) == FALSE) {
+      if (grepl(s, l$rhs.terms[k]) == FALSE) {
         minimum <- 1
       } else {
-        minimum <- as.integer(sub(s, "\\1", env$rhs.terms[k], perl = TRUE))
+        minimum <- as.integer(sub(s, "\\1", l$rhs.terms[k], perl = TRUE))
       }
       
       # extract maximum argument
       s <- "(?:timecov\\(.*maximum\\s*=\\s*)(\\d+)(?:.*\\))"
-      if (grepl(s, env$rhs.terms[k]) == FALSE) {
-        maximum <- env$time.steps
+      if (grepl(s, l$rhs.terms[k]) == FALSE) {
+        maximum <- l$time.steps
       } else {
-        maximum <- as.integer(sub(s, "\\1", env$rhs.terms[k], perl = TRUE))
+        maximum <- as.integer(sub(s, "\\1", l$rhs.terms[k], perl = TRUE))
       }
       
       # extract transform argument
       s <- "(?:timecov\\(.*transform\\s*=\\s*)(.+?)(?:(?:,|\\)$)]*.*)"
-      if (grepl(s, env$rhs.terms[k]) == FALSE) {
+      if (grepl(s, l$rhs.terms[k]) == FALSE) {
         transform <- function(t) 1 + (0 * t) + (0 * t^2)
       } else {
-        transform <- eval(parse(text = sub(s, "\\1", env$rhs.terms[k], 
+        transform <- eval(parse(text = sub(s, "\\1", l$rhs.terms[k], 
             perl = TRUE)))
       }
       
       # process dependent list of networks
       if (is.null(x)) {
-        covariate <- env[["networks"]]
+        covariate <- l[["networks"]]
         onlytime <- TRUE
       } else {
         onlytime <- FALSE
@@ -300,29 +298,29 @@ tergmprepare <- function(formula, offset = TRUE, blockdiag = FALSE,
         maximum = maximum, transform = transform, onlytime = onlytime)
       
       # re-introduce as edgecov and name of model term including brackets
-      env[[label]] <- tc
+      l[[label]] <- tc
       if (blockdiag == TRUE) {
-        env$rhs.terms[k] <- paste0("edgecov(", label, ")")
+        l$rhs.terms[k] <- paste0("edgecov(", label, ")")
       } else {
-        env$rhs.terms[k] <- paste0("edgecov(", label, "[[i]])")
+        l$rhs.terms[k] <- paste0("edgecov(", label, "[[i]])")
       }
-      env$covnames <- c(env$covnames, label)
+      l$covnames <- c(l$covnames, label)
     }
   }
-  env$covnames <- c("networks", env$covnames)
+  l$covnames <- c("networks", l$covnames)
   
   # fix different lengths of DV/covariate lists due to temporal dependencies
-  lengths <- sapply(env$covnames, function(cn) length(env[[cn]]))
+  lengths <- sapply(l$covnames, function(cn) length(l[[cn]]))
   mn <- max(lengths)
   if (length(table(lengths)) > 1) {
     mn <- min(lengths)
-    env$time.steps <- mn
-    for (i in 1:length(env$covnames)) {
-      cn <- env$covnames[[i]]
-      l <- env[[cn]]
-      difference <- length(l) - mn
+    l$time.steps <- mn
+    for (i in 1:length(l$covnames)) {
+      cn <- l$covnames[[i]]
+      ll <- l[[cn]]
+      difference <- length(ll) - mn
       if (difference > 0) {
-        env[[cn]] <- l[(difference + 1):length(l)]
+        l[[cn]] <- ll[(difference + 1):length(ll)]
       }
     }
   }
@@ -331,16 +329,16 @@ tergmprepare <- function(formula, offset = TRUE, blockdiag = FALSE,
   
   # determine and report initial dimensions of networks and covariates
   if (verbose == TRUE) {
-    if (length(env$covnames) > 1) {
-      dimensions <- lapply(lapply(env$covnames, function(x) env[[x]]), 
+    if (length(l$covnames) > 1) {
+      dimensions <- lapply(lapply(l$covnames, function(x) l[[x]]), 
           function(y) sapply(y, function(z) dim(as.matrix(z))))
-      rownames(dimensions[[1]]) <- paste(env$lhs.original, c("(row)", "(col)"))
+      rownames(dimensions[[1]]) <- paste(l$lhs.original, c("(row)", "(col)"))
       for (i in 2:length(dimensions)) {
-        rownames(dimensions[[i]]) <- c(paste(env$covnames[i], "(row)"), 
-            paste(env$covnames[i], "(col)"))
+        rownames(dimensions[[i]]) <- c(paste(l$covnames[i], "(row)"), 
+            paste(l$covnames[i], "(col)"))
       }
       dimensions <- do.call(rbind, dimensions)
-      colnames(dimensions) <- paste0("t=", t.start:t.end) #1:length(env$networks))
+      colnames(dimensions) <- paste0("t=", t.start:t.end) #1:length(l$networks))
       message("\nInitial dimensions of the network and covariates:")
       print(dimensions)
     } else {
@@ -349,36 +347,36 @@ tergmprepare <- function(formula, offset = TRUE, blockdiag = FALSE,
   }
   
   # determine whether covariate dimensions need to be automatically adjusted
-  env$auto.adjust <- FALSE
-  if (length(env$covnames) > 1) {
+  l$auto.adjust <- FALSE
+  if (length(l$covnames) > 1) {
     # check number of rows and columns
-    nr <- lapply(lapply(env$covnames, function(x) env[[x]]), 
+    nr <- lapply(lapply(l$covnames, function(x) l[[x]]), 
         function(y) sapply(y, function(z) nrow(as.matrix(z))))
     nr <- do.call(rbind, nr)
-    nc <- lapply(lapply(env$covnames, function(x) env[[x]]), 
+    nc <- lapply(lapply(l$covnames, function(x) l[[x]]), 
         function(y) sapply(y, function(z) ncol(as.matrix(z))))
     nc <- do.call(rbind, nc)
     for (i in 1:ncol(nr)) {
       if (length(unique(nr[, i])) > 1) {
-        env$auto.adjust <- TRUE
+        l$auto.adjust <- TRUE
       }
     }
     for (i in 1:ncol(nc)) {
       if (length(unique(nc[, i])) > 1) {
-        env$auto.adjust <- TRUE
+        l$auto.adjust <- TRUE
       }
     }
-    if (verbose == TRUE && env$auto.adjust == TRUE) {
+    if (verbose == TRUE && l$auto.adjust == TRUE) {
        message(paste("\nDimensions differ across networks within time steps."))
     }
     # check if labels are present
-    if (env$auto.adjust == TRUE) {
-      for (i in 1:length(env$covnames)) {
-        for (t in 1:env$time.steps) {
-          if (is.null(rownames(as.matrix(env[[env$covnames[i]]][[t]]))) || 
-              is.null(colnames(as.matrix(env[[env$covnames[i]]][[t]])))) {
+    if (l$auto.adjust == TRUE) {
+      for (i in 1:length(l$covnames)) {
+        for (t in 1:l$time.steps) {
+          if (is.null(rownames(as.matrix(l[[l$covnames[i]]][[t]]))) || 
+              is.null(colnames(as.matrix(l[[l$covnames[i]]][[t]])))) {
             stop(paste0("The dimensions of the covariates differ, but ", 
-                "covariate '", env$covnames[i], 
+                "covariate '", l$covnames[i], 
                 " does not have node labels at t = ", t, 
                 ". Automatic adjustment of dimensions is therefore not ", 
                 "possible."))
@@ -387,13 +385,13 @@ tergmprepare <- function(formula, offset = TRUE, blockdiag = FALSE,
       }
     }
     # check if there are different labels despite identical dimensions
-    if (env$auto.adjust == FALSE) {
-      for (t in 1:env$time.steps) {
+    if (l$auto.adjust == FALSE) {
+      for (t in 1:l$time.steps) {
         rlabels.i <- list()
         clabels.i <- list()
-        for (i in 1:length(env$covnames)) {
-          rlabels.i[[i]] <- rownames(as.matrix(env[[env$covnames[i]]][[t]]))
-          clabels.i[[i]] <- colnames(as.matrix(env[[env$covnames[i]]][[t]]))
+        for (i in 1:length(l$covnames)) {
+          rlabels.i[[i]] <- rownames(as.matrix(l[[l$covnames[i]]][[t]]))
+          clabels.i[[i]] <- colnames(as.matrix(l[[l$covnames[i]]][[t]]))
         }
         rlabels.i <- do.call(rbind, rlabels.i)
         clabels.i <- do.call(rbind, clabels.i)
@@ -401,7 +399,7 @@ tergmprepare <- function(formula, offset = TRUE, blockdiag = FALSE,
         if (!is.null(rlabels.i)) {
           for (j in 1:ncol(rlabels.i)) {
             if (length(unique(rlabels.i[, j])) > 1) {
-              env$auto.adjust <- TRUE
+              l$auto.adjust <- TRUE
               flag <- TRUE
               break
             }
@@ -410,7 +408,7 @@ tergmprepare <- function(formula, offset = TRUE, blockdiag = FALSE,
         if (!is.null(clabels.i)) {
           for (j in 1:ncol(clabels.i)) {
             if (length(unique(clabels.i[, j])) > 1) {
-              env$auto.adjust <- TRUE
+              l$auto.adjust <- TRUE
               flag <- TRUE
               break
             }
@@ -423,7 +421,7 @@ tergmprepare <- function(formula, offset = TRUE, blockdiag = FALSE,
       }
     }
   }
-  if (verbose == TRUE && env$auto.adjust == TRUE) {
+  if (verbose == TRUE && l$auto.adjust == TRUE) {
     message("Trying to auto-adjust the dimensions of the networks. ", 
         "If this fails, provide conformable matrices or network objects.")
   } else if (verbose == TRUE) {
@@ -433,28 +431,28 @@ tergmprepare <- function(formula, offset = TRUE, blockdiag = FALSE,
   # do mutual adjustment of networks and covariates at each time step
   structzero.df <- data.frame(label = character(), time = integer(), 
       object = character(), where = character())
-  if (length(env$covnames) > 0 && env$auto.adjust == TRUE) {
-    for (i in 1:env$time.steps) {
-      for (j in 1:length(env$covnames)) {
-        for (k in 1:length(env$covnames)) {
+  if (length(l$covnames) > 0 && l$auto.adjust == TRUE) {
+    for (i in 1:l$time.steps) {
+      for (j in 1:length(l$covnames)) {
+        for (k in 1:length(l$covnames)) {
           if (j != k) {
-            nw.j <- env[[env$covnames[j]]][[i]]
+            nw.j <- l[[l$covnames[j]]][[i]]
             rn.j <- rownames(as.matrix(nw.j))
             cn.j <- colnames(as.matrix(nw.j))
             nr.j <- nrow(as.matrix(nw.j))
             nc.j <- ncol(as.matrix(nw.j))
-            nw.k <- env[[env$covnames[k]]][[i]]
+            nw.k <- l[[l$covnames[k]]][[i]]
             rn.k <- rownames(as.matrix(nw.k))
             cn.k <- colnames(as.matrix(nw.k))
             nr.k <- nrow(as.matrix(nw.k))
             nc.k <- ncol(as.matrix(nw.k))
             if (is.null(rn.j) || is.null(cn.j)) {
               stop(paste0("Missing row or column labels in object '", 
-                  env$covnames[j], "'. Provide row and column ", 
+                  l$covnames[j], "'. Provide row and column ", 
                   "labels for all networks and covariates."))
             } else if (is.null(rn.k) || is.null(cn.k)) {
               stop(paste0("Missing row or column labels in object '", 
-                  env$covnames[k], "'. Provide row and column ", 
+                  l$covnames[k], "'. Provide row and column ", 
                   "labels for all networks and covariates."))
             } else {
               if (is.null(rn.j) && !is.null(rn.k) && nr.j == nr.k) {
@@ -470,26 +468,26 @@ tergmprepare <- function(formula, offset = TRUE, blockdiag = FALSE,
                   rownames(nw.k) <- rn.j
                 }
               } else if ((is.null(rn.k) || is.null(rn.j)) && nr.j != nr.k) {
-                stop(paste0("Object '", env$covnames[j], 
-                    "' is incompatible with object '", env$covnames[k], 
+                stop(paste0("Object '", l$covnames[j], 
+                    "' is incompatible with object '", l$covnames[k], 
                     "' at t = ", i, "."))
               }
               # adjust j to k
               nw.j.labels <- adjust(nw.j, nw.k, remove = FALSE, 
                   value = 1, returnlabels = TRUE)
               nw.j <- adjust(nw.j, nw.k, remove = FALSE, value = 1)
-              env[[env$covnames[j]]][[i]] <- nw.j
+              l[[l$covnames[j]]][[i]] <- nw.j
               ro <- nw.j.labels$added.row
               co <- nw.j.labels$added.col
               if (length(ro) > 0) {
                 ro <- data.frame(label = ro, time = rep(i, length(ro)), 
-                    object = rep(env$covnames[j], length(ro)), 
+                    object = rep(l$covnames[j], length(ro)), 
                     where = rep("row", length(ro)))
                 structzero.df <- rbind(structzero.df, ro)
               }
               if (length(co) > 0) {
                 co <- data.frame(label = co, time = rep(i, length(co)), 
-                    object = rep(env$covnames[j], length(co)), 
+                    object = rep(l$covnames[j], length(co)), 
                     where = rep("col", length(co)))
                 structzero.df <- rbind(structzero.df, co)
               }
@@ -497,18 +495,18 @@ tergmprepare <- function(formula, offset = TRUE, blockdiag = FALSE,
               nw.k.labels <- adjust(nw.k, nw.j, remove = FALSE, 
                   value = 1, returnlabels = TRUE)
               nw.k <- adjust(nw.k, nw.j, remove = FALSE, value = 1)
-              env[[env$covnames[k]]][[i]] <- nw.k
+              l[[l$covnames[k]]][[i]] <- nw.k
               ro <- nw.k.labels$added.row
               co <- nw.k.labels$added.col
               if (length(ro) > 0) {
                 ro <- data.frame(label = ro, time = rep(i, length(ro)), 
-                    object = rep(env$covnames[j], length(ro)), 
+                    object = rep(l$covnames[j], length(ro)), 
                     where = rep("row", length(ro)))
                 structzero.df <- rbind(structzero.df, ro)
               }
               if (length(co) > 0) {
                 co <- data.frame(label = co, time = rep(i, length(co)), 
-                    object = rep(env$covnames[j], length(co)), 
+                    object = rep(l$covnames[j], length(co)), 
                     where = rep("col", length(co)))
                 structzero.df <- rbind(structzero.df, co)
               }
@@ -520,27 +518,27 @@ tergmprepare <- function(formula, offset = TRUE, blockdiag = FALSE,
   }
   
   # check whether all dimensions are cross-sectionally conformable now
-  nr.net <- sapply(env$networks, function(x) nrow(as.matrix(x)))
-  for (i in 1:length(env$covnames)) {
-    nr <- sapply(env[[env$covnames[i]]], function(x) {
+  nr.net <- sapply(l$networks, function(x) nrow(as.matrix(x)))
+  for (i in 1:length(l$covnames)) {
+    nr <- sapply(l[[l$covnames[i]]], function(x) {
       nrow(as.matrix(x))
     })
-    for (j in 1:env$time.steps) {
+    for (j in 1:l$time.steps) {
       if (nr[j] != nr.net[j]) {
-        stop(paste0("Covariate object '", env$covnames[i], 
+        stop(paste0("Covariate object '", l$covnames[i], 
             "' does not have the same number of rows as the dependent ", 
             "network at time step ", j, "."))
       }
     }
   }
-  nc.net <- sapply(env$networks, function(x) ncol(as.matrix(x)))
-  for (i in 1:length(env$covnames)) {
-    nc <- sapply(env[[env$covnames[i]]], function(x) {
+  nc.net <- sapply(l$networks, function(x) ncol(as.matrix(x)))
+  for (i in 1:length(l$covnames)) {
+    nc <- sapply(l[[l$covnames[i]]], function(x) {
       ncol(as.matrix(x))
     })
-    for (j in 1:env$time.steps) {
+    for (j in 1:l$time.steps) {
       if (nc[j] != nc.net[j]) {
-        stop(paste0("Covariate object '", env$covnames[i], 
+        stop(paste0("Covariate object '", l$covnames[i], 
             "' does not have the same number of columns as the dependent ", 
             "network at time step ", j, "."))
       }
@@ -549,19 +547,19 @@ tergmprepare <- function(formula, offset = TRUE, blockdiag = FALSE,
   
   # reporting
   if (verbose == TRUE) {
-    if (env$auto.adjust == TRUE) {
+    if (l$auto.adjust == TRUE) {
       sz.row <- unique(structzero.df[structzero.df$where == "row", -3])
-      szrownum <- numeric(length(env$networks))
-      for (i in 1:length(env$networks)) {
+      szrownum <- numeric(length(l$networks))
+      for (i in 1:length(l$networks)) {
         szrownum[i] <- nrow(sz.row[sz.row$time == i, ])
       }
       sz.col <- unique(structzero.df[structzero.df$where == "col", -3])
-      szcolnum <- numeric(length(env$networks))
-      for (i in 1:length(env$networks)) {
+      szcolnum <- numeric(length(l$networks))
+      for (i in 1:length(l$networks)) {
         szcolnum[i] <- nrow(sz.col[sz.col$time == i, ])
       }
-      totrow <- sapply(env$networks, function(x) nrow(as.matrix(x)))
-      totcol <- sapply(env$networks, function(x) ncol(as.matrix(x)))
+      totrow <- sapply(l$networks, function(x) nrow(as.matrix(x)))
+      totcol <- sapply(l$networks, function(x) ncol(as.matrix(x)))
       if (offset == TRUE) {
         dimensions <- rbind(totrow, totcol, szrownum, szcolnum, 
             totrow - szrownum, totcol - szcolnum)
@@ -596,25 +594,30 @@ tergmprepare <- function(formula, offset = TRUE, blockdiag = FALSE,
     }
   }
   
+  l$nvertices <- sapply(l$networks, function(x) c(nrow(as.matrix(x)), 
+      ncol(as.matrix(x))))
+  rownames(l$nvertices) <- c("row", "col")
+  colnames(l$nvertices) <- paste0("t=", t.start:t.end)
+  
   # create list of offset matrices (required both for offset and node removal)
-  env$offsmat <- list()
-  for (i in 1:env$time.steps) {
-    mat <- matrix(0, nrow = nrow(as.matrix(env$networks[[i]])), 
-        ncol = ncol(as.matrix(env$networks[[i]])))
-    rownames(mat) <- rownames(as.matrix(env$networks[[i]]))
-    colnames(mat) <- colnames(as.matrix(env$networks[[i]]))
-    env$offsmat[[i]] <- mat
+  l$offsmat <- list()
+  for (i in 1:l$time.steps) {
+    mat <- matrix(0, nrow = nrow(as.matrix(l$networks[[i]])), 
+        ncol = ncol(as.matrix(l$networks[[i]])))
+    rownames(mat) <- rownames(as.matrix(l$networks[[i]]))
+    colnames(mat) <- colnames(as.matrix(l$networks[[i]]))
+    l$offsmat[[i]] <- mat
   }
   if (nrow(structzero.df) > 0) {
     for (i in 1:nrow(structzero.df)) {
       if (structzero.df$where[i] == "row") {
-        index <- which(rownames(env$offsmat[[structzero.df$time[i]]]) == 
+        index <- which(rownames(l$offsmat[[structzero.df$time[i]]]) == 
             structzero.df$label[i])
-        env$offsmat[[structzero.df$time[i]]][index, ] <- 1
+        l$offsmat[[structzero.df$time[i]]][index, ] <- 1
       } else {
-        index <- which(colnames(env$offsmat[[structzero.df$time[i]]]) == 
+        index <- which(colnames(l$offsmat[[structzero.df$time[i]]]) == 
             structzero.df$label[i])
-        env$offsmat[[structzero.df$time[i]]][, index] <- 1
+        l$offsmat[[structzero.df$time[i]]][, index] <- 1
       }
     }
   }
@@ -622,99 +625,102 @@ tergmprepare <- function(formula, offset = TRUE, blockdiag = FALSE,
   # offset preparation or node removal for MPLE
   if (offset == TRUE) {
     # add offset to formula and reassemble formula
-    env$rhs.terms[length(env$rhs.terms) + 1] <- "offset(edgecov(offsmat[[i]]))"
+    l$rhs.terms[length(l$rhs.terms) + 1] <- "offset(edgecov(offsmat[[i]]))"
     rhs.operators[length(rhs.operators) + 1] <- "+"
   } else {
     # delete nodes with structural zeros
-    if (env$auto.adjust == TRUE) {
-      env$offsmat <- suppressMessages(handleMissings(env$offsmat, na = 1, 
+    if (l$auto.adjust == TRUE) {
+      l$offsmat <- suppressMessages(handleMissings(l$offsmat, na = 1, 
           method = "remove"))
-      for (j in 1:length(env$covnames)) {
-        env[[env$covnames[j]]] <- adjust(env[[env$covnames[j]]], env$offsmat)
+      for (j in 1:length(l$covnames)) {
+        l[[l$covnames[j]]] <- adjust(l[[l$covnames[j]]], l$offsmat)
       }
     }
   }
   
   # determine and report initial dimensions of networks and covariates
-  if (verbose == TRUE && length(env$covnames) > 1) {
-    dimensions <- lapply(lapply(env$covnames, function(x) env[[x]]), 
+  if (verbose == TRUE && length(l$covnames) > 1) {
+    dimensions <- lapply(lapply(l$covnames, function(x) l[[x]]), 
         function(y) sapply(y, function(z) dim(as.matrix(z))))
-    rownames(dimensions[[1]]) <- paste(env$lhs.original, c("(row)", "(col)"))
+    rownames(dimensions[[1]]) <- paste(l$lhs.original, c("(row)", "(col)"))
     for (i in 2:length(dimensions)) {
-      rownames(dimensions[[i]]) <- c(paste(env$covnames[i], "(row)"), 
-          paste(env$covnames[i], "(col)"))
+      rownames(dimensions[[i]]) <- c(paste(l$covnames[i], "(row)"), 
+          paste(l$covnames[i], "(col)"))
     }
     dimensions <- do.call(rbind, dimensions)
-    colnames(dimensions) <- paste0("t=", t.start:t.end) #1:length(env$networks))
+    colnames(dimensions) <- paste0("t=", t.start:t.end) #1:length(l$networks))
     message("\nDimensions of the network and covariates after adjustment:")
     print(dimensions)
   }
   
   # assemble formula
-  rhs <- env$rhs.terms[1]
+  rhs <- l$rhs.terms[1]
   if (length(rhs.operators) > 0) {
     for (i in 1:length(rhs.operators)) {
-      rhs <- paste(rhs, rhs.operators[i], env$rhs.terms[i + 1])
+      rhs <- paste(rhs, rhs.operators[i], l$rhs.terms[i + 1])
     }
   }
   f <- paste(lhs, tilde, rhs)
-  env$form <- as.formula(f, env = env)
+  l$form <- as.formula(f, env = environment())  # l$form <- as.formula(f)
   
   # for mtergm estimation using MCMC: create block-diagonal matrices
   if (blockdiag == TRUE) {
-    if (env$bipartite == TRUE) {
+    if (l$bipartite == TRUE) {
       stop(paste("MCMC estimation is currently only supported for one-mode", 
           "networks. Use the btergm function instead."))
     }
     # also save formula without time indices for ergm estimation
-    env$form <- update.formula(env$form, networks ~ .)
-    env$form <- paste(deparse(env$form), collapse = "")
-    env$form <- paste(env$form, "+ offset(edgecov(offsmat))")
-    env$form <- as.formula(env$form, env = env)
+    l$form <- update.formula(l$form, networks ~ .)
+    l$form <- paste(deparse(l$form), collapse = "")
+    l$form <- paste(l$form, "+ offset(edgecov(offsmat))")
+    l$form <- as.formula(l$form, env = environment())
     # make covariates block-diagonal
-    if (length(env$covnames) > 1) {
-      for (j in 2:length(env$covnames)) {
-        env[[env$covnames[j]]] <- as.matrix(Matrix::bdiag(lapply(
-            env[[env$covnames[j]]], as.matrix)))
+    if (length(l$covnames) > 1) {
+      for (j in 2:length(l$covnames)) {
+        l[[l$covnames[j]]] <- as.matrix(Matrix::bdiag(lapply(
+            l[[l$covnames[j]]], as.matrix)))
       }
     }
     # create block-diagonal offset matrix and merge with existing offsmat term
-    env$offsmat <- as.matrix(Matrix::bdiag(env$offsmat))  # make block-diagonal
-    bdoffset <- lapply(env$networks, as.matrix)
+    l$offsmat <- as.matrix(Matrix::bdiag(l$offsmat))  # make block-diagonal
+    bdoffset <- lapply(l$networks, as.matrix)
     for (i in 1:length(bdoffset)) {
       bdoffset[[i]][, ] <- 1
     }
     bdoffset <- as.matrix((Matrix::bdiag(bdoffset) - 1) * -1)  # off-diagonal
-    env$offsmat <- env$offsmat + bdoffset
+    l$offsmat <- l$offsmat + bdoffset
     rm(bdoffset)
-    env$offsmat[env$offsmat > 0] <- 1
+    l$offsmat[l$offsmat > 0] <- 1
     # make dependent network block-diagonal
-    if (class(env$networks[[1]]) == "network") {  # network
-      attrnames <- network::list.vertex.attributes(env$networks[[1]])
+    if (class(l$networks[[1]]) == "network") {  # network
+      attrnames <- network::list.vertex.attributes(l$networks[[1]])
       attributes <- list()
-      for (i in 1:length(env$networks)) {
+      for (i in 1:length(l$networks)) {
         attrib <- list()
         for (j in 1:length(attrnames)) {
-          attrib[[j]] <- network::get.vertex.attribute(env$networks[[i]], 
+          attrib[[j]] <- network::get.vertex.attribute(l$networks[[i]], 
               attrnames[j])
         }
         attributes[[i]] <- attrib
-        env$networks[[i]] <- as.matrix(env$networks[[i]])
+        l$networks[[i]] <- as.matrix(l$networks[[i]])
       }
-      env$networks <- network::network(as.matrix(Matrix::bdiag(env$networks)), 
-          directed = env$directed, bipartite = env$bipartite)
+      l$networks <- network::network(as.matrix(Matrix::bdiag(l$networks)), 
+          directed = l$directed, bipartite = l$bipartite)
       for (i in 1:length(attrnames)) {  # collate attributes and merge back in
         attrib <- unlist(lapply(attributes, function(x) x[[i]]))
-        network::set.vertex.attribute(env$networks, attrnames[i], attrib)
+        network::set.vertex.attribute(l$networks, attrnames[i], attrib)
       }
     } else {  # matrix
-      env$networks <- network::network(as.matrix(Matrix::bdiag(env$networks)), 
-          directed = env$directed, bipartite = env$bipartite)
+      l$networks <- network::network(as.matrix(Matrix::bdiag(l$networks)), 
+          directed = l$directed, bipartite = l$bipartite)
     }
     if (verbose == TRUE) {
       cat("\n")  # to get a blank line before the MCMC MLE output starts
     }
   }
-  return(env)  # return the environment with all the data
+  # convert formula back to character object because environment invalid anyway
+  form3 <- paste(deparse(l$form[[3]]), collapse = "")  # for long formulae
+  form3 <- gsub("\\s+", " ", form3)
+  l$form <- paste(deparse(l$form[[2]]), deparse(l$form[[1]]), form3)
+  return(l)  # return the environment with all the data
 }
-
